#ifndef CX__H
#define CX__H

class CX // Using class RE defined in <re.h>
{
public:
  RE x, y;
  // Construction and conversion operations
  CX () { } // Constructeur par dfaut
  CX (RE a)                            : x(a), y(0) { }
  CX (RE a,          RE b)             : x(a), y(b) { }
  CX (int a,         int b = 0)        : x(a), y(b) { }
  CX (long a,        long b = 0)       : x(a), y(b) { }
  CX (float a,       float b = 0)      : x(a), y(b) { }
  CX (double a,      double b = 0)     : x(a), y(b) { }
  inline CX (const char *a, const char *b) { x << a; y << b; }
  inline CX (const char *ab) { *this << ab; }
  ~CX() { }
  // Fundamental comparison operators
  friend bool operator ==( const CX& a, const CX& b );
  friend bool IsZero( const CX& a );
  friend bool IsOne ( const CX& a );
  friend bool IsReal( const CX& a );
  // Fundamental arithmetic operators
  friend CX operator -( const CX& a );
  friend CX   inv( const CX& a ); // returns the inverse
  friend CX  conj( const CX& a ); // returns the conjugate
  friend RE   dot( const CX& a, const CX& b ); // returns the dot product
  friend RE  norm( const CX& a ); // returns the norm
  CX& operator +=( const CX& a );
  CX& operator *=( const CX& a );
  CX& operator -=( const CX& a );
  CX& operator /=( const CX& a ) { *this *= inv(a); return *this; }
  // Fundamental output functions
  friend void operator <<(char*& s, const CX& c);
  friend void operator <<(CX& c, const char* s);
  // Adding a product may be faster with a redefinition of this function
  void addprod (const CX& a, const CX& b);
};

// Fundamental comparison operators
inline bool IsZero ( const CX& a ) { return IsZero(a.x) && IsZero(a.y); }
inline bool IsOne  ( const CX& a ) { return IsOne (a.x) && IsZero(a.y); }
inline bool IsReal ( const CX& a ) { return IsZero(a.y); }
inline bool operator ==( const CX& a, const CX& b )
  { return (a.x == b.x) && (a.y == b.y); }
// Fundamental arithmetic operators
inline CX operator -( const CX& a ) { return CX( - a.x, - a.y ); }
inline CX conj ( const CX& a ) { return CX( a.x, -a.y ); }
inline RE dot ( const CX& a, const CX& b ) { return a.x * b.x + a.y * b.y; }
inline CX inv ( const CX& a )
{
  RE norme = norm(a);
  if (IsZero(norme)) { Error("CX: division by zero"); throw norme; }
  return CX( a.x / norme, - a.y / norme );
}
inline CX& CX::operator +=( const CX& a ) { x += a.x; y += a.y; return *this; }
inline CX& CX::operator *=( const CX& a )
{
  RE r = a.x * x  -  a.y * y;
  y = a.x * y  +  a.y * x;
  x = r;
  return *this;
}
// Fundamental output functions
// [cx-io.c] void operator <<( char*& s, const CX& c );
// [cx-io.c] void operator <<( CX& c, const char* s );


// -------- Redefinitions of additional methods --------
inline RE norm ( const CX& a ) { return sqr(a.x) + sqr(a.y); }
inline CX operator + (const CX& a, const CX& b) { return CX(a.x+b.x, a.y+b.y); }
inline CX operator - (const CX& a, const CX& b) { return CX(a.x-b.x, a.y-b.y); }
inline CX operator * (const CX& a, const CX& b)
{
  return CX( a.x * b.x  -  a.y * b.y,
             a.x * b.y  +  a.y * b.x );
}
inline CX& CX::operator -=( const CX& a ) { x -= a.x; y -= a.y; return *this; }
inline bool operator !=( const CX& a, const CX& b )
  { return (a.x != b.x) || (a.y != b.y); }
inline void CX::addprod(const CX& a, const CX& b)
{
  RE r;
  mul(r, a.x,b.x); add(x, x,r);
  mul(r, a.y,b.y); sub(x, x,r);
  mul(r, a.x,b.y); add(y, y,r);
  mul(r, a.y,b.x); add(y, y,r);
}

// -------- Definitions from fundamental methods --------
inline CX operator / (const CX& a, const CX& b) { return a * inv(b); }
inline CX sqr ( const CX& a ) { return a * a; }
inline CX unit ( const CX& a )
  { if (!IsZero(a)) return a/CX(sqrt(norm(a)),0); else return a; }
inline ostream& operator<<(ostream& s, const CX& a)
  { char *str; str << a; s << str; free(str); return s; }
inline CX pow( const CX& c, const unsigned long exposant )
{
  CX res = 1;
  unsigned long k = exposant;
  CX a = c;
  for( ; k > 1; k >>= 1) { if (k&1) res *= a; a *= a; }
  if (k) res *= a;
  return res;
}

#endif
